import io.github.jaredmdobson.concentus.*
import kotlinx.coroutines.*
import java.util.concurrent.atomic.AtomicInteger
import javax.print.Doc
import kotlin.concurrent.fixedRateTimer

/**
 * Created by LoliBall on 2024/8/31 5:26.
 * https://github.com/WhichWho
 */

//val SERVER_HOST = "zedo.top"
val SERVER_HOST = "127.0.0.1"
val SERVER_PORT = 7860
val ECHO_CANCELLATION = false // 回声消除
val RAW_PCM_PASS = true // 不进行opus压缩pcm直传

// 2.5, 5,   10,  20,  40,   60 ms
// 120, 240, 480, 960, 1920, 2880
val PACKET_SAMPLES = 2880 // 48000Hz * 60ms = 2880
val SAMPLE_RATE = 48000 // 8000、12000、16000、24000、48000
val CHANNELS = 1
val BUFFER_SIZE = PACKET_SAMPLES * CHANNELS * 2 // * 2是short | 48000Hz * 2(short) * 60ms = 5760B
val AUDIO_FORMAT = AudioConvert.getAudioFormat(SAMPLE_RATE, CHANNELS)

val OPUS_BITRATE = OpusConstants.OPUS_AUTO // 500 - 300000
val OPUS_COMPLEXITY = 10 // 0: 快,质量差 10: 慢,质量好
val OPUS_SIGNAL_TYPE = OpusSignal.OPUS_SIGNAL_VOICE
val OPUS_MODE = OpusMode.MODE_HYBRID


// 下面是统计用的
val compressRateOpus = AtomicInteger()
val compressRatePcm = AtomicInteger()

fun main(args: Array<String>) {
    val scope = CoroutineScope(Dispatchers.IO)
    scope.launch {
        Server.main(emptyArray())
    }
    scope.launch {
        delay(1000) // 等服务器起来
        Client.main(emptyArray())
    }
    fixedRateTimer(period = 1000) {
        "compress rate: %.2f%%".format(compressRateOpus.get() * 100f / compressRatePcm.get()).loge()
        compressRateOpus.set(0)
        compressRatePcm.set(0)
    }
}

val encodeOpusBuffer = mutableMapOf<OpusEncoder, ByteArray>()
val encodePcmShortBuffer = mutableMapOf<OpusEncoder, ShortArray>()

// 1275是opus单帧包最大值
fun OpusEncoder.encode(pcm: ByteArray): DataPack {
    if (RAW_PCM_PASS) {
        return DataPack(System.currentTimeMillis(), pcm.copyOf())
    }
    val opusBuffer = encodeOpusBuffer[this] ?: ByteArray(1275).also { encodeOpusBuffer[this] = it }
    val shortBuffer = encodePcmShortBuffer[this] ?: ShortArray(pcm.size / 2).also { encodePcmShortBuffer[this] = it }
    AudioConvert.bytes2shorts(pcm, shortBuffer)
    val bytesEncoded = encode(shortBuffer, 0, PACKET_SAMPLES, opusBuffer, 0, opusBuffer.size)
    compressRatePcm.addAndGet(pcm.size)
    compressRateOpus.addAndGet(bytesEncoded)
    return DataPack(System.currentTimeMillis(), opusBuffer.sliceArray(0..<bytesEncoded))
}

val decodePcmShortBuffer = mutableMapOf<OpusDecoder, ShortArray>()
fun OpusDecoder.decode(pack: DataPack, pcm: ByteArray) {
    if (RAW_PCM_PASS) {
        pack.opus.copyInto(pcm)
        return
    }
    val shortBuffer = decodePcmShortBuffer[this] ?: ShortArray(pcm.size / 2).also { decodePcmShortBuffer[this] = it }
    decode(pack.opus, 0, pack.opus.size, shortBuffer, 0, PACKET_SAMPLES, false)
    AudioConvert.shorts2bytes(shortBuffer, pcm)
}

fun <T> T.loge() = also { println(this) }